/**
 * PETALS - PETALS Services Platform. Copyright (c) 2008 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.se.eip.patterns;

import java.net.URI;
import java.util.ArrayList;
import java.util.List;
import java.util.SortedMap;
import java.util.TreeMap;

import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.transform.TransformerException;
import javax.xml.transform.dom.DOMSource;
import javax.xml.xpath.XPathExpressionException;

import org.ow2.petals.component.framework.Constants.MEPConstants;
import org.ow2.petals.component.framework.api.configuration.ConfigurationExtensions;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.util.SourceHelper;
import org.ow2.petals.component.framework.util.XMLUtil;
import org.ow2.petals.se.eip.EIPConstants;
import org.ow2.petals.se.eip.ExchangeContext;
import org.w3c.dom.Document;

/**
 * A Router pattern
 * 
 * @author alouis
 * @author ofabre
 * @author Christophe HAMERLING - eBM WebSourcing
 * @since 2.2
 * 
 */
public class DynamicRouter extends AbstractPattern {

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#process(org.ow2.petals
     * .component.framework.util.Exchange,
     * org.ow2.petals.se.eip.ExchangeContext)
     */
    @Override
    public void process(Exchange exchange, ExchangeContext context) throws MessagingException,
            PEtALSCDKException {
        List<Consumes> consumesList = context.getSUConsumes(exchange.getEndpoint());

        List<String> conditions = retrieveConditions(context.getExtensions());

        // 0. there must be numberOfConditions+2 consumes defined in the SU
        int count = consumesList.size();
        if (count != conditions.size() + 2) {
            throw new PEtALSCDKException(
                    "The Service Unit must define (numberOfConditions+2) Consumes sections for the pattern "
                            + this.getPatternName());
        }

        Exchange originalExchange = exchange;
        Document inMessage = originalExchange.getInMessageContentAsDocument();
        originalExchange.getInMessage().setContent(new DOMSource(inMessage));

        // 1. the first' consumes' is the targeted service used to evaluate the
        // condition
        Consumes conditionService = consumesList.get(0);
        Exchange conditionExchange = PatternHelper.createExchange(conditionService,
                MEPConstants.IN_OUT_PATTERN, originalExchange.getInMessage(), context);

        if (conditionExchange.getOperation() == null) {
            conditionExchange.setOperation(originalExchange.getOperation());
        }
        logger.info(getPatternName()
                + " send a copy, for evaluation, of the IN message to ...\n - itf="
                + conditionService.getInterfaceName() + "\n - srv="
                + conditionService.getServiceName() + "\n - ep ="
                + conditionService.getEndpointName());
        context.sendSync(conditionExchange);

        // 2. process the result of the Service to evaluate
        // find the Consumes to call depending on the xpath conditions
        NormalizedMessage receivedConditionMessage = PatternHelper.processSentExchange(
                conditionExchange, originalExchange, context);
        if (receivedConditionMessage != null) {
            Document conditionMessage = SourceHelper.createDocument(receivedConditionMessage
                    .getContent());
            // 2.1 evaluate the result with conditions to retrieve the Consumes
            // to call
            // (default Consumes is used if no conditions is True)
            int targetedConsumesListIndentifier;
            try {
                targetedConsumesListIndentifier = retrieveTargetedConsumesListIdentifier(
                        conditions, conditionMessage, context);
            } catch (XPathExpressionException e) {
                throw new PEtALSCDKException(e);
            }

            // 2.2 forward the original exchange to the Service
            Consumes consumedService = consumesList.get(targetedConsumesListIndentifier);
            Exchange routedExchange = PatternHelper.createExchange(consumedService, MEPConstants
                    .valueOf(originalExchange.getPattern()), originalExchange.getInMessage(),
                    context);

            if (routedExchange.getOperation() == null) {
                routedExchange.setOperation(originalExchange.getOperation());
            }

            logSend(consumedService, routedExchange);
            context.sendSync(routedExchange);

            // 2.3 process the response
            NormalizedMessage receivedMessage = PatternHelper.processSentExchange(routedExchange,
                    originalExchange, context);
            // if a response has been returned, set it on the original
            // exchange
            // otherwise, the' processSentExchange()' has manage the
            // exchange
            if (receivedMessage != null) {
                originalExchange.setOutMessage(receivedMessage);
            }
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.AbstractPattern#getPatternName()
     */
    @Override
    protected String getPatternName() {
        return "DynamicRouter";
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#validateMEP(java.net.URI)
     */
    @Override
    protected boolean validateMEP(URI mep) {
        return MEPConstants.IN_ONLY_PATTERN.equals(mep)
                || MEPConstants.ROBUST_IN_ONLY_PATTERN.equals(mep)
                || MEPConstants.IN_OUT_PATTERN.equals(mep);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.Pattern#init()
     */
    public void init() {
    }

    /**
     * 
     * @param conditions
     * @param inMessage
     * @param context
     * @return
     * @throws XPathExpressionException
     */
    private int retrieveTargetedConsumesListIdentifier(List<String> conditions, Document inMessage,
            ExchangeContext context) throws XPathExpressionException {
        // Choose the default one first
        int targetedConsumesListIndentifier = conditions.size() + 1;
        for (int i = 0; i < conditions.size(); i++) {
            // evaluate the IN message
            try {
                String evaluation = XMLUtil.createStringFromDOMDocument(inMessage);
                logger.info(getPatternName() + " document to evaluate : " + evaluation);
            } catch (TransformerException e) {
                logger.warning(getPatternName() + " document to evaluate may be empty");
            }
            logger.info(getPatternName() + " evaluates condition #" + (i + 1) + " : "
                    + conditions.get(i));
            if (PatternHelper.evaluateXPathCondition(conditions.get(i), inMessage.getFirstChild(),
                    context)) {
                logger.info(getPatternName() + " evaluation is OK");
                targetedConsumesListIndentifier = i + 1;
                break;
            }
        }
        if (targetedConsumesListIndentifier == conditions.size() + 1) {
            logger.info(getPatternName()
                    + " no condition is satisfied, default Service will be called");
        }
        return targetedConsumesListIndentifier;
    }

    /**
     * 
     * @param extensions
     * @return
     */
    private List<String> retrieveConditions(ConfigurationExtensions extensions) {
        List<String> conditions = new ArrayList<String>();
        SortedMap<Integer, String> orderedConditions = new TreeMap<Integer, String>();

        for (String param : extensions.keySet()) {
            if (EIPConstants.TEST.equals(param)) {
                conditions.add(extensions.get(param));
                break;
            } else if (param.startsWith(EIPConstants.TEST)) {
                String indexString = param.substring(EIPConstants.TEST.length());
                orderedConditions.put(Integer.valueOf(indexString), extensions.get(param));
            }
        }

        if (conditions.size() == 0) {
            conditions.addAll(orderedConditions.values());
        }

        return conditions;
    }
}
