/**
 * PETALS - PETALS Services Platform. Copyright (c) 2007 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.se.eip.patterns;

import java.util.Set;

import javax.jbi.JBIException;
import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.security.auth.Subject;
import javax.xml.transform.Source;
import javax.xml.transform.stream.StreamSource;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.ow2.petals.component.framework.Constants.MEPConstants;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.jbidescriptor.generated.MEPType;
import org.ow2.petals.component.framework.util.InputStreamForker;
import org.ow2.petals.se.eip.EIPException;
import org.ow2.petals.se.eip.ExchangeContext;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * 
 * @author alouis
 * 
 */
public class PatternHelper {

	private static final XPath XPATH;
	static {
		XPATH = XPathFactory.newInstance().newXPath();
	}


	/**
	 * Process the result of a sent exchange.<br>
	 * - if the exchange contains an OUT response, return it and closed the sent
	 * exchange<br>
	 * - if the exchange contains a FAULT, set it on the original exchange and
	 * return nothing, sent exchange is closed<br>
	 * - if the exchange is ERROR, set it on the original exchange and return
	 * nothing<br>
	 * - if the exchange is DONE, return nothing<br>
	 * 
	 * @param sentExchange
	 * @param originalExchange
	 * @param context
	 * @return the OUT message of the sent exchange, if any. When null, the
	 *         original exchange has been updated
	 * @throws JBIException
	 */
	public static final NormalizedMessage processSentExchange(
			Exchange sentExchange, Exchange originalExchange,
			ExchangeContext context) throws MessagingException {
		NormalizedMessage receivedMessage = null;
		context.getLogger().finest(
				"Process exchange received from : srv="
						+ sentExchange.getEndpoint().getServiceName()
						+ " / ep="
						+ sentExchange.getEndpoint().getEndpointName());
		if (sentExchange.isActiveStatus()) {
			if (sentExchange.getFault() != null) {
				context.getLogger().finest("a fault has been returned");
				if (originalExchange.isInOnlyPattern()) {
					context
							.getLogger()
							.finest(
									"orginal exchange is InOnly, the fault can not be returned.");
				} else {
					context.getLogger().finest(
							"set the fault on the original Exchange.");
					// fault on piped exchange, set it on the original exchange
					// do not return a message
					Fault fault = originalExchange.createFault();
					fault.setContent(sentExchange.getFault().getContent());
					for (Object name : sentExchange.getFault()
							.getPropertyNames()) {
						fault.setProperty((String) name, sentExchange
								.getFault().getProperty((String) name));
					}
					for (Object name : sentExchange.getFault()
							.getAttachmentNames()) {
						fault.addAttachment((String) name, sentExchange
								.getFault().getAttachment((String) name));
					}
					originalExchange.setFault(fault);
				}
			} else {
				// out message received, return it
				context.getLogger().finest("an Out message has been returned.");
				receivedMessage = sentExchange.getOutMessage();
				for (Object name : sentExchange.getOutMessage()
						.getPropertyNames()) {
					receivedMessage.setProperty((String) name, sentExchange
							.getOutMessage().getProperty((String) name));
				}
				for (Object name : sentExchange.getOutMessage()
						.getAttachmentNames()) {
					receivedMessage.addAttachment((String) name, sentExchange
							.getOutMessage().getAttachment((String) name));
				}
			}
			// close the sent exchange
			sentExchange.setStatus(ExchangeStatus.DONE);
			context.send(sentExchange);
		} else if (sentExchange.isErrorStatus()) {
			// error on the sent exchange, set it on the original exchange
			// do not return a message
			context
					.getLogger()
					.finest(
							"an Error status has been received, set the original Exchange status to Error.");
			originalExchange.setError(sentExchange.getError());
		} else if (sentExchange.isDoneStatus()) {
			// sent exchange return nothing (done)
			// do not return a message
			context.getLogger().finest(
					"a Done status has been received, return nothing.");
		}
		return receivedMessage;
	}

	/**
	 * Process the result of an exchange.<br>
	 * - if the exchange contains an OUT response, return it and close the
	 * exchange<br>
	 * - if the exchange contains a FAULT, return the fault and close the
	 * exchange<br>
	 * - if the exchange is ERROR, return an EIPException and close the exchange<br>
	 * - if the exchange is DONE, return nothing<br>
	 * 
	 * @param sentExchange
	 *            the exchange sent
	 * @param context
	 *            The context used
	 * @return the OUT message of the exchange or null
	 * @throws MessagingException
	 * @throws EIPException
	 */
	public static final NormalizedMessage processSentExchange(
			Exchange sentExchange, ExchangeContext context)
			throws MessagingException, EIPException {
		NormalizedMessage receivedMessage = null;
		context.getLogger().fine(
				"Process exchange received from the service "
						+ sentExchange.getEndpoint().getServiceName()
						+ " in the endpoint "
						+ sentExchange.getEndpoint().getEndpointName());
		if (sentExchange.isActiveStatus()) {
			if (sentExchange.getFault() != null) {
				context.getLogger().info("a fault is returned.");
				receivedMessage = sentExchange.getFault();
				receivedMessage
						.setContent(sentExchange.getFault().getContent());
				for (Object name : sentExchange.getFault().getPropertyNames()) {
					receivedMessage.setProperty((String) name, sentExchange
							.getFault().getProperty((String) name));
				}
				for (Object name : sentExchange.getFault().getAttachmentNames()) {
					receivedMessage.addAttachment((String) name, sentExchange
							.getFault().getAttachment((String) name));
				}
			} else {
				// out message received, return it
				context.getLogger().info("an out message is returned.");
				receivedMessage = sentExchange.getOutMessage();
				for (Object name : sentExchange.getOutMessage()
						.getPropertyNames()) {
					receivedMessage.setProperty((String) name, sentExchange
							.getOutMessage().getProperty((String) name));
				}
				for (Object name : sentExchange.getOutMessage()
						.getAttachmentNames()) {
					receivedMessage.addAttachment((String) name, sentExchange
							.getOutMessage().getAttachment((String) name));
				}
			}
			// close the exchange
			sentExchange.setStatus(ExchangeStatus.DONE);
			context.send(sentExchange);
		} else if (sentExchange.isErrorStatus()) {
			// error on exchange, return an EIPException
			// do not return a message
			context
					.getLogger()
					.fine(
							"the exchange status is Error, set the original Exchange status to Error.");
			throw new EIPException(sentExchange.getError());
		} else if (sentExchange.isDoneStatus()) {
			// service return nothing, exchange is closed
			// do not return a message
			context.getLogger()
					.fine("the exchange status is Done, do nothing.");
		}
		return receivedMessage;
	}

	/**
	 * Create an exchange for the given Consumes services. Set the IN message
	 * with a <em>copy</em> of the given message
	 * 
	 * @param consumes
	 * @param mep
	 * @param originalMessage
	 * @return
	 * @throws MessagingException
	 * @throws PEtALSCDKException
	 */
	public static final Exchange createExchange(Consumes consumes,
			MEPConstants mep, NormalizedMessage originalMessage,
			ExchangeContext context) throws MessagingException,
			PEtALSCDKException {
		Exchange tmpExchange = context.createConsumeExchange(consumes, mep);
		context.getLogger().finest("tmpExchange created");
		NormalizedMessage tmpMessage = tmpExchange.getInMessage();
		context.getLogger().finest("tmpMessage created");
		copy(originalMessage, tmpMessage);
		context.getLogger().finest(
				"tmpExchange initialyzed with the originalExchange values");
		return tmpExchange;

	}

	/**
	 * Create a consume exchange for the given Consumes services.
	 * 
	 * @param consumes
	 *            The consumes used
	 * @param context
	 *            the context used
	 * @return the Exchange created
	 * @throws MessagingException
	 * @throws PEtALSCDKException
	 */
	public static final Exchange createConsumeExchange(Consumes consumes,
			ExchangeContext context) throws MessagingException,
			PEtALSCDKException {
		MEPConstants mep = PatternHelper.retrievePattern(consumes.getMep());
		if (mep == null) {
			throw new PEtALSCDKException(
					"The consumes doesn't have a valid Message Exchange Pattern : unable to generate a correct Exchange");
		}
		Exchange exchange = context.createConsumeExchange(consumes, mep);
		context.getLogger().finest("exchange created");
		return exchange;
	}

	/**
	 * Copies message's content, security, properties and attachments in
	 * another.
	 * 
	 * @param source
	 *            The NormalizedMessage source (copied)
	 * @param target
	 *            The NormalizedMessage target (pasted)
	 * @throws MessagingException
	 *             error while setting attachment
	 * @throws PEtALSCDKException
	 *             error while using PEtALS object
	 */
	public static final void copy(final NormalizedMessage source,
			final NormalizedMessage target) throws MessagingException,
			PEtALSCDKException {
		copy(source.getContent(), source.getSecuritySubject(), source, target);
	}

	/**
	 * Copies message's properties and attachments, and content and security
	 * provided, in another.
	 * 
	 * @param content
	 *            the Source content (copied)
	 * @param securitySubject
	 *            the security Subject (copied)
	 * @param source
	 *            The NormalizedMessage source (properties and attachments
	 *            copied)
	 * @param target
	 *            The NormalizedMessage target (pasted)
	 * @throws MessagingException
	 *             error while setting attachment
	 * @throws PEtALSCDKException
	 *             error while using PEtALS object
	 */
	public static final void copy(Source content, Subject securitySubject,
			NormalizedMessage source, NormalizedMessage target)
			throws MessagingException, PEtALSCDKException {
		// setting content
		target.setContent(content);
		// setting security context
		target.setSecuritySubject(securitySubject);
		// setting properties
		copyProperties(source, target);
		// setting attachments if any
		copyAttachments(source, target);
	}

	/**
	 * Retrieve the MEP from the extensions.
	 * 
	 * @param extensions
	 * @return mep specified in extensions
	 * @throws MessagingException
	 *             MEP pattern specified in consumes extension isn't valid or if
	 *             no MEP pattern specified in extensions.
	 */
	public static final MEPConstants retrievePattern(MEPType patternShortcut)
			throws MessagingException {
		MEPConstants mep = null;

		if (patternShortcut != null) {
			if (MEPType.IN_ONLY.equals(patternShortcut)) {
				mep = MEPConstants.IN_ONLY_PATTERN;
			} else if (MEPType.ROBUST_IN_ONLY.equals(patternShortcut)) {
				mep = MEPConstants.ROBUST_IN_ONLY_PATTERN;
			} else if (MEPType.IN_OUT.equals(patternShortcut)) {
				mep = MEPConstants.IN_OUT_PATTERN;
			} else if (MEPType.IN_OPTIONAL_OUT.equals(patternShortcut)) {
				mep = MEPConstants.IN_OPTIONAL_OUT_PATTERN;
			}
		}

		return mep;
	}

	/**
	 * 
	 * @param condition
	 * @param message
	 * @param context
	 * @return
	 * @throws XPathExpressionException
	 */
	public static final boolean evaluateXPathCondition(String expression,
			final Node message, ExchangeContext context)
			throws XPathExpressionException {
		String result = evaluateXPathExpression(expression, message, context);
		return result.equals("true");
	}

	/**
	 * 
	 * @param expression
	 * @param message
	 * @param context
	 * @return
	 * @throws XPathExpressionException
	 */
	public static final String evaluateXPathExpression(String expression,
			final Node message, ExchangeContext context)
			throws XPathExpressionException {
		String result = "";
		final XPathExpression xpathExpression = XPATH.compile(expression);
		result = xpathExpression.evaluate(message);
		return result;
	}

	/**
	 * 
	 * @param expression
	 *            The XPath expression to compile
	 * @param message
	 *            The Node to evaluate
	 * @return The NodeList evaluated by XPath under the message
	 * @throws XPathExpressionException
	 */
	public static final NodeList evaluateXPathExpression(String expression,
			final Node message) throws XPathExpressionException {
		XPathExpression xpathExpression = XPATH.compile(expression);
		return (NodeList) xpathExpression.evaluate(message,
				XPathConstants.NODESET);
	}

	

	/**
	 * Handle the streaming issue.
	 * 
	 * @param streamSource
	 * @throws MessagingException
	 */
	public static final InputStreamForker handleStreamSource(
			NormalizedMessage message) throws MessagingException {
		InputStreamForker inMessageAsStream = null;
		if (message.getContent() instanceof StreamSource) {
			inMessageAsStream = new InputStreamForker(((StreamSource) message
					.getContent()).getInputStream());
			Source inSource = new StreamSource(inMessageAsStream
					.getInputStreamOne());
			message.setContent(inSource);
		}

		return inMessageAsStream;
	}

	/**
	 * Handle the streaming issue.
	 * 
	 * @param streamSource
	 * @throws MessagingException
	 */
	public static final InputStreamForker handleStreamSource(Fault fault)
			throws MessagingException {
		InputStreamForker faultMessageAsStream = null;
		if (fault.getContent() instanceof StreamSource) {
			faultMessageAsStream = new InputStreamForker(((StreamSource) fault
					.getContent()).getInputStream());
			Source faultSource = new StreamSource(faultMessageAsStream
					.getInputStreamOne());
			fault.setContent(faultSource);
		}

		return faultMessageAsStream;
	}

	/**
	 * copy the attachments from
	 * 
	 * @param sourceMessage
	 *            message containing the attachments
	 * @param destMessage
	 * 
	 * @throws PEtALSCDKException
	 * @throws MessagingException
	 */
	@SuppressWarnings("unchecked")
	public static void copyAttachments(NormalizedMessage sourceMessage,
			NormalizedMessage destMessage) throws PEtALSCDKException,
			MessagingException {
		Set<String> attNames = sourceMessage.getAttachmentNames();
		for (String string : attNames) {
			destMessage.addAttachment(string, sourceMessage
					.getAttachment(string));
		}
	}

	/**
	 * copy the properties
	 * 
	 * @param destMessage
	 *            message containing the properties
	 * @param destMessage
	 * 
	 * @throws PEtALSCDKException
	 * @throws MessagingException
	 */
	@SuppressWarnings("unchecked")
	public static void copyProperties(NormalizedMessage sourceMessage,
			NormalizedMessage destMessage) throws PEtALSCDKException,
			MessagingException {
		Set<String> attNames = sourceMessage.getPropertyNames();
		for (String string : attNames) {
			destMessage.setProperty(string, sourceMessage.getProperty(string));
		}
	}
}
