/**
 * PETALS - PETALS Services Platform. Copyright (c) 2007 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.se.eip.patterns;

import java.net.URI;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;

import javax.jbi.JBIException;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.transform.Source;
import javax.xml.transform.stream.StreamSource;

import org.ow2.petals.component.framework.Constants.MEPConstants;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.util.InputStreamForker;
import org.ow2.petals.component.framework.util.SourceHelper;
import org.ow2.petals.component.framework.util.XMLUtil;
import org.ow2.petals.se.eip.EIPConstants;
import org.ow2.petals.se.eip.ExchangeContext;
import org.w3c.dom.Document;
import org.w3c.dom.Node;

/**
 * Aggregate all Consumes services defined in the service unit. <br>
 * Return the aggregated results.<br>
 * <br>
 * - Accept InOut<br>
 * - If a Fault or Error occurs on a call, stop the process and return the Fault
 * or Error Note : The exchange will be sent back automatically by the framework
 * 
 * @author alouis, chamerling - eBM WebSourcing
 * 
 */
public class ScatterGather extends AbstractPattern {


    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#process(org.ow2.petals
     * .component.framework.util.Exchange,
     * org.ow2.petals.se.eip.ExchangeContext)
     */
    public void process(Exchange exchange, ExchangeContext context) throws PEtALSCDKException {
        Exchange originalExchange = exchange;

        List<Consumes> consumesList = context.getSUConsumes(originalExchange.getEndpoint());

        if (consumesList.size() == 0) {
            throw new PEtALSCDKException(
                    "The Service Unit must define at least 1 Consumes sections for the pattern "
                            + this.getPatternName());
        }

        List<Exchange> sentExchanges = sendAsyncExchanges(consumesList, exchange, context);

        // when messages have been sent, wait for the responses from all
        // services
        try {
            if (sentExchanges != null) {
                receiveResponse(sentExchanges, exchange, context);
            }
        } catch (MessagingException e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
            try {
                exchange.setFault(e);
            } catch (MessagingException ex) {
                logger.log(Level.SEVERE, ex.getMessage(), ex);
            }
        }

    }

    /**
     * Send a copy of the exchange's IN message to all the services to consume.
     * Return the list of the Exchanges sent asynchronously. If an error occurs,
     * set a Fault on the original exchange, and return null TODO if an exchange
     * fails, close the previous sent exchange...
     * 
     * @param consumesList
     * @param exchange
     * @param context
     * @return the list of the exchanges sent async, or null if a problem
     *         occurred
     */
    protected List<Exchange> sendAsyncExchanges(List<Consumes> consumesList, Exchange exchange,
            ExchangeContext context) {

        List<Exchange> sentExchanges = new ArrayList<Exchange>();
        try {
            NormalizedMessage inMessage = exchange.getInMessage();

            // Handle the Streaming issue
            InputStreamForker inMessageAsStream = PatternHelper.handleStreamSource(inMessage);
            // send all exchanges asynchronously
            for (Consumes consumedService : consumesList) {
                if (inMessageAsStream != null) {
                    // Handle the Streaming issue
                    Source inSource = new StreamSource(inMessageAsStream.getInputStreamOne());
                    inMessageAsStream = new InputStreamForker(inMessageAsStream.getInputStreamTwo());
                    inMessage.setContent(inSource);
                }

                MEPConstants mep = PatternHelper.retrievePattern(consumedService.getMep());
                if (mep == null) {
                    mep = MEPConstants.valueOf(exchange.getPattern());
                }
                Exchange tmpExchange = PatternHelper.createExchange(consumedService, mep,
                        inMessage, context);

                logSend(consumedService, tmpExchange);
                context.sendAsync(tmpExchange);
                sentExchanges.add(tmpExchange);
            }
        } catch (JBIException e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
            try {
                exchange.setFault(e);
            } catch (MessagingException ex) {
                logger.log(Level.SEVERE, ex.getMessage(), ex);
            }
            sentExchanges = null;
        }
        return sentExchanges;
    }

    /**
     * aggregates all the responses (xml content, properties and attachments)
     * @param sentExchanges
     * @param originalExchange
     * @param context
     * @return
     * @throws MessagingException 
     */
    protected void receiveResponse(List<Exchange> sentExchanges, Exchange originalExchange,
            ExchangeContext context) throws MessagingException {
        // wait for all exchanges responses

        // build an aggregate document for the result
        Document aggregateDocument = XMLUtil.createDocument();
        Node aggregateRootElement = aggregateDocument.createElementNS(EIPConstants.NAMESPACE,EIPConstants.RESULT_ELEMENT);
        aggregateDocument.appendChild(aggregateRootElement);

        try {
            for (Exchange sentExchange : sentExchanges) {
                Exchange aggregateExchange = context.accept(sentExchange);
                // process the response. If fault or error, stop the
                // process and return the fault
                // the' processSentExchange()' has set the fault
                // on the original exchange
                NormalizedMessage receivedMessage = PatternHelper.processSentExchange(
                        aggregateExchange, originalExchange, context);
                if (receivedMessage != null) {
                    // aggregate the XML content, attachments and properties
                    aggregateContent(receivedMessage, aggregateDocument);
                    originalExchange.setOutMessageContent(aggregateDocument);
                    PatternHelper.copyAttachments(receivedMessage, originalExchange.getOutMessage());
                    PatternHelper.copyProperties(receivedMessage, originalExchange.getOutMessage());
                } else {
                    // The fault has already been set the original exchange
                }
            }
        } catch (JBIException e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
            try {
                originalExchange.setFault(e);
            } catch (MessagingException e1) {
                logger.log(Level.SEVERE, e1.getMessage(), e1);
            }
        } catch (InterruptedException e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
            try {
                originalExchange.setFault(e);
            } catch (MessagingException e1) {
                logger.log(Level.SEVERE, e1.getMessage(), e1);
            }
        }
    }

    /**
     * Aggregate the item XML content into the global AggregateDocument
     * 
     * @param receivedMessage
     *            message containing the xml content to append
     * @param aggregateDocument
     *            the global document in which the item is appended
     * @throws PEtALSCDKException
     */
    protected void aggregateContent(NormalizedMessage receivedMessage, Document aggregateDocument)
            throws PEtALSCDKException {
        // save all the root NS
        Source src = receivedMessage.getContent();
        Document itemDocument = SourceHelper.createDocument(src);
        Node itemElement = itemDocument.getFirstChild();

        aggregateDocument.adoptNode(itemElement);
        aggregateDocument.getFirstChild().appendChild(itemElement);
    }

   
    
    
    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.Pattern#init()
     */
    public void init() {
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.AbstractPattern#getPatternName()
     */
    @Override
    protected String getPatternName() {
        return "ScatterGather";
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#validateMEP(java.net.URI)
     */
    @Override
    protected boolean validateMEP(URI mep) {
        return MEPConstants.IN_OUT_PATTERN.equals(mep);
    }
}
