/**
 * The Splitter EIP pattern
 */
package org.ow2.petals.se.eip.patterns;

import java.net.URI;
import java.util.ArrayList;
import java.util.List;

import javax.jbi.messaging.Fault;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.transform.dom.DOMSource;
import javax.xml.xpath.XPathExpressionException;

import org.ow2.petals.component.framework.Constants.MEPConstants;
import org.ow2.petals.component.framework.api.configuration.ConfigurationExtensions;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.se.eip.EIPConstants;
import org.ow2.petals.se.eip.EIPException;
import org.ow2.petals.se.eip.ExchangeContext;
import org.w3c.dom.Document;
import org.w3c.dom.NodeList;

/**
 * @author Adrien LOUIS - EBMWebSourcing
 * @author fgardes
 * 
 */
public class Splitter extends AbstractPattern {

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.AbstractPattern#getPatternName()
     */
    @Override
    protected String getPatternName() {
        return "Splitter";
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#process(org.ow2.petals
     * .component.framework.api.message.Exchange,
     * org.ow2.petals.se.eip.ExchangeContext)
     */
    @Override
    public void process(Exchange exchange, ExchangeContext context) throws MessagingException,
            PEtALSCDKException {
        List<Consumes> consumesList = null;
        List<Exchange> splitExchangeList = null;
        Consumes consumes = null;
        String splitterPath = null;
        NormalizedMessage receivedMessage = null;
        boolean fault = false;

        if (!exchange.isDoneStatus()) {
            consumesList = context.getSUConsumes(exchange.getEndpoint());
            if (consumesList == null || consumesList.size() != 1) {
                throw new PEtALSCDKException(
                        "The Service Unit must define only one consumes section");
            }
            consumes = consumesList.get(0);
            if (consumes == null) {
                throw new PEtALSCDKException(
                        "The Service Unit must define a valid consumes section");
            }
            splitterPath = getSplitterPath(context.getExtensions());
            if (splitterPath == null || splitterPath.length() == 0) {
                throw new PEtALSCDKException(
                        "The Service Unit must define a splitter-element section");
            }
            logger.finest("splitter element:" + splitterPath);
            // retrieve the document only once to avoid the Streaming issue
            Document inMessageContent = exchange.getInMessageContentAsDocument();
            try {
                NodeList nodeList = PatternHelper.evaluateXPathExpression(splitterPath,
                        inMessageContent);
                logger.finest("node list generated with " + nodeList.getLength() + " elements");
                splitExchangeList = new ArrayList<Exchange>(nodeList.getLength());
                logger.finest("split exchange list generated");
                for (int i = 0; i < nodeList.getLength(); i++) {
                    Exchange splitExchange = PatternHelper.createConsumeExchange(consumes, context);
                    if (splitExchange.getOperation() == null) {
                        splitExchange.setOperation(exchange.getOperation());
                    }
                    PatternHelper.copy(new DOMSource(nodeList.item(i)), exchange.getInMessage()
                            .getSecuritySubject(), exchange.getInMessage(), splitExchange
                            .getInMessage());
                    logSend(consumes, splitExchange);
                    context.sendAsync(splitExchange);
                    splitExchangeList.add(splitExchange);
                    logger.info("split exchange " + splitExchange.getExchangeId() + " sent");
                }
                for (int i = 0; i < splitExchangeList.size() && !fault; i++) {
                    // process the response
                    receivedMessage = PatternHelper.processSentExchange(context
                            .accept(splitExchangeList.get(i)), context);
                    logger.finest("split exchange analyzed");
                    if (receivedMessage != null) {
                        if (receivedMessage instanceof Fault && exchange.isRobustInOnlyPattern()) {
                            exchange.setFault((Fault) receivedMessage);
                            fault = true;
                            exchange.setDoneStatus();
                            logger.finest("fault received");
                        }
                    }
                }
                // reset the content of the in Message to avoid Streaming issue
                exchange.getInMessage().setContent(new DOMSource(inMessageContent));
            } catch (XPathExpressionException e) {
                throw new PEtALSCDKException(e);
            } catch (EIPException e) {
                exchange.setError((Exception) e.getCause());
                exchange.setErrorStatus();
            } catch (InterruptedException e) {
                exchange.setError(e);
                exchange.setErrorStatus();
            }
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#validateMEP(java.net.URI)
     */
    @Override
    protected boolean validateMEP(URI mep) {
        return MEPConstants.IN_ONLY_PATTERN.equals(mep)
                || MEPConstants.ROBUST_IN_ONLY_PATTERN.equals(mep);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.Pattern#init()
     */
    public void init() {
    }

    /**
     * get the splitter element prefix into the extensions
     * 
     * @param extensions
     *            the ConfigurationExtensions used
     * @return The splitter element, or null if not found
     */
    private String getSplitterPath(ConfigurationExtensions extensions) {
        for (String param : extensions.keySet()) {
            if (EIPConstants.PATH.equals(param)) {
                return extensions.get(param);
            }
        }
        return null;
    }

}
