/**
 * The Splitter EIP pattern
 */

package org.ow2.petals.se.eip.patterns;

import java.net.URI;
import java.util.ArrayList;
import java.util.List;

import javax.jbi.messaging.Fault;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.transform.Source;
import javax.xml.transform.dom.DOMSource;
import javax.xml.xpath.XPathExpressionException;

import org.ow2.petals.component.framework.Constants.MEPConstants;
import org.ow2.petals.component.framework.api.configuration.ConfigurationExtensions;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.util.SourceHelper;
import org.ow2.petals.component.framework.util.XMLUtil;
import org.ow2.petals.se.eip.EIPConstants;
import org.ow2.petals.se.eip.EIPException;
import org.ow2.petals.se.eip.ExchangeContext;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * @author Adrien LOUIS - EBMWebSourcing
 * @author fgardes
 * 
 */
public class SplitterGather extends AbstractPattern {

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.AbstractPattern#getPatternName()
     */
    @Override
    protected String getPatternName() {
        return "SplitterGather";
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#process(org.ow2.petals
     * .component.framework.api.message.Exchange,
     * org.ow2.petals.se.eip.ExchangeContext)
     */
    @Override
    public void process(Exchange originalExchange, ExchangeContext context)
            throws MessagingException, PEtALSCDKException {
        List<NormalizedMessage> outMessageList = null;
        List<Consumes> consumesList = null;
        List<Exchange> splitExchangeList = null;
        Consumes consumes = null;
        String splitterPath = null;
        NormalizedMessage receivedMessage = null;
        boolean fault = false;

        if (!originalExchange.isDoneStatus()) {
            consumesList = context.getSUConsumes(originalExchange.getEndpoint());
            if (consumesList == null || consumesList.size() != 1) {
                throw new PEtALSCDKException(
                        "The Service Unit must define only one consumes section");
            }
            consumes = consumesList.get(0);
            if (consumes == null) {
                throw new PEtALSCDKException(
                        "The Service Unit must define a valid consumes section");
            }
            splitterPath = getSplitterPath(context.getExtensions());
            if (splitterPath == null || splitterPath.length() == 0) {
                throw new PEtALSCDKException(
                        "The Service Unit must define a splitter-element section");
            }
            logger.finest("splitter element:" + splitterPath);
            // retrieve the document only once to avoid the Streaming issue
            Document inMessageContent = originalExchange.getInMessageContentAsDocument();
            try {
                NodeList nodeList = PatternHelper.evaluateXPathExpression(splitterPath,
                        inMessageContent);
                logger.finest("node list generated with " + nodeList.getLength() + " elements");
                outMessageList = new ArrayList<NormalizedMessage>(nodeList.getLength());
                logger.finest("out message list generated");
                splitExchangeList = new ArrayList<Exchange>(nodeList.getLength());
                logger.finest("split exchange list generated");
                for (int i = 0; i < nodeList.getLength(); i++) {
                    Exchange splitExchange = PatternHelper.createConsumeExchange(consumes, context);
                    if (splitExchange.getOperation() == null) {
                        splitExchange.setOperation(originalExchange.getOperation());
                    }
                    PatternHelper.copy(new DOMSource(nodeList.item(i)), originalExchange
                            .getInMessage().getSecuritySubject(), originalExchange.getInMessage(),
                            splitExchange.getInMessage());
                    logSend(consumes, splitExchange);
                    context.sendAsync(splitExchange);
                    splitExchangeList.add(splitExchange);
                    logger.info("split exchange " + splitExchange.getExchangeId() + " sent");
                }
                for (int i = 0; i < splitExchangeList.size() && !fault; i++) {
                    // process the response
                    receivedMessage = PatternHelper.processSentExchange(context
                            .accept(splitExchangeList.get(i)), context);
                    logger.finest("split exchange analyzed");
                    if (receivedMessage != null) {
                        if (receivedMessage instanceof Fault) {
                            originalExchange.setFault((Fault) receivedMessage);
                            fault = true;
                        } else {
                            outMessageList.add(receivedMessage);
                            logger.finest("message received");
                        }
                    }
                }
                // reset the content of the in Message to avoid Streaming issue
                originalExchange.getInMessage().setContent(new DOMSource(inMessageContent));
                // manage out , fault and done differently for each mep
                if (!fault) {
                    // put out message to content
                    originalExchange.getOutMessage().setContent(
                            concatNormalizedMesssageContent(outMessageList));
                    if (outMessageList != null) {
                        for (int i = 0; i < outMessageList.size(); i++) {
                            receivedMessage = outMessageList.get(i);
                            // put each properties
                            for (Object name : receivedMessage.getPropertyNames()) {
                                originalExchange.getOutMessage().setProperty((String) name,
                                        receivedMessage.getProperty((String) name));
                            }
                            // put each attachments
                            for (Object name : receivedMessage.getAttachmentNames()) {
                                originalExchange.getOutMessage().addAttachment((String) name,
                                        receivedMessage.getAttachment((String) name));
                            }

                        }
                    }
                }
            } catch (XPathExpressionException e) {
                throw new PEtALSCDKException(e);
            } catch (EIPException e) {
                originalExchange.setError((Exception) e.getCause());
                originalExchange.setErrorStatus();
            } catch (InterruptedException e) {
                originalExchange.setError(e);
                originalExchange.setErrorStatus();
            }

        }
    }

    /**
     * Create a document with the root node provided by the context
     * 
     * @return a Document with the root node provided by the context
     * @throws PEtALSCDKException
     */
    private Document createResponse() throws PEtALSCDKException {

        Document document = XMLUtil.createDocument();
        Element rootElement = document.createElementNS(EIPConstants.NAMESPACE,
                EIPConstants.RESULT_ELEMENT);
        document.appendChild(rootElement);
        return document;
    }

    /**
     * Build the aggregate source from a list of normalized messages with the
     * root node provided by the context.
     * 
     * @param normalizedMessageList
     *            The List of NormalizedMessage used
     * @return a Source with the content of each valid message content
     * @throws PEtALSCDKException
     */
    private Source concatNormalizedMesssageContent(List<NormalizedMessage> normalizedMessageList)
            throws PEtALSCDKException {
        NormalizedMessage normalizedMessage = null;
        Document aggregateDocument = createResponse();
        if (normalizedMessageList != null) {
            for (int i = 0; i < normalizedMessageList.size(); i++) {
                normalizedMessage = normalizedMessageList.get(i);
                Document outDocument = SourceHelper.createDocument(normalizedMessage.getContent());
                Node node = aggregateDocument.adoptNode(outDocument.getDocumentElement());
                aggregateDocument.getDocumentElement().appendChild(node);
            }
        }
        return SourceHelper.createSource(aggregateDocument);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#validateMEP(java.net.URI)
     */
    @Override
    protected boolean validateMEP(URI mep) {
        return MEPConstants.IN_OUT_PATTERN.equals(mep);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.Pattern#init()
     */
    public void init() {
    }

    /**
     * get the splitter element prefix into the extensions
     * 
     * @param extensions
     *            the ConfigurationExtensions used
     * @return The splitter element, or null if not found
     */
    private String getSplitterPath(ConfigurationExtensions extensions) {
        for (String param : extensions.keySet()) {
            if (EIPConstants.PATH.equals(param)) {
                return extensions.get(param);
            }
        }
        return null;
    }

}
